//
//  ViewController.swift
//  AztecDemoSwift
//
//  Created by PrimeSoft Polska on 09.03.2016.
//  Copyright © 2016 PrimeSoft Polska. All rights reserved.
//

import UIKit
import AVFoundation
import CoreVideo
import CoreMedia

class ViewController: UIViewController, AVCaptureVideoDataOutputSampleBufferDelegate {
    
    let MAX_THREADS = 4
    let USE_60_FPS = false
    let USE_15_FPS = false
    
    enum cameraState {
        case NORMAL
        case CAMERA
        case CAMERA_DECODING
        case DECODE_DISPLAY
    }
    
    private var device: AVCaptureDevice!
    private var captureSession: AVCaptureSession!
    private var previewLayer: AVCaptureVideoPreviewLayer!
    private var focusTimer: NSTimer!
    
    var state: cameraState = cameraState.NORMAL
    var activeThreads:Int!
    var availableThreads:Int!
    
    var tokenOnce: dispatch_once_t = 0
    let maskLayer = MaskLayer();
    
    //MARK: - View life cycle
    override func viewDidLoad() {
        super.viewDidLoad()
        // Do any additional setup after loading the view, typically from a nib.        
    }
    
    override func viewWillAppear(animated: Bool) {
        super.viewWillAppear(animated);
        let auth  = checkCameraAuthorizationStatus()
        if !auth.flag {
            if let vc = UIStoryboard(name: "Main", bundle: nil).instantiateViewControllerWithIdentifier("\(MessageViewController.self)") as? MessageViewController {
                vc.message = auth.message
                
                dispatch_async(dispatch_get_main_queue(), { 
                    self.presentViewController(vc, animated: true, completion: nil)
                })
            }
        }else {
            self.initCapture();
            self.startScanning();   
            
            let frame = self.view.bounds
            self.previewLayer.frame = frame;
            
            self.view.layer.addSublayer(maskLayer);
            
            updateMaskLayer(frame)
        }
    }
    
    override func viewWillDisappear(animated: Bool) {
        super.viewWillDisappear(animated);
        
        stopScanning();
        deinitCapture();
        maskLayer.removeFromSuperlayer()
    }
    
    override func willRotateToInterfaceOrientation(toInterfaceOrientation: UIInterfaceOrientation, duration: NSTimeInterval ){
        
        // UIView.setAnimationsEnabled(false);
        
        if toInterfaceOrientation == UIInterfaceOrientation.LandscapeLeft{
            self.previewLayer.connection.videoOrientation = AVCaptureVideoOrientation.LandscapeLeft;
            let frame = CGRectMake(0, 0, max(self.view.frame.size.width,self.view.frame.size.height), min(self.view.frame.size.width,self.view.frame.size.height));
            self.previewLayer.frame = frame
            updateMaskLayer(frame)
            
        }
        
        if toInterfaceOrientation == UIInterfaceOrientation.LandscapeRight{
            self.previewLayer.connection.videoOrientation = AVCaptureVideoOrientation.LandscapeRight;
            let frame = CGRectMake(0, 0, max(self.view.frame.size.width,self.view.frame.size.height), min(self.view.frame.size.width,self.view.frame.size.height));
            self.previewLayer.frame = frame;
            updateMaskLayer(frame)
        }
        
        if toInterfaceOrientation == UIInterfaceOrientation.Portrait{
            self.previewLayer.connection.videoOrientation = AVCaptureVideoOrientation.Portrait;
            let frame = CGRectMake(0, 0, min(self.view.frame.size.width,self.view.frame.size.height), max(self.view.frame.size.width,self.view.frame.size.height));
            self.previewLayer.frame = frame
            updateMaskLayer(frame)
        }
        
        if toInterfaceOrientation == UIInterfaceOrientation.PortraitUpsideDown{
            self.previewLayer.connection.videoOrientation = AVCaptureVideoOrientation.PortraitUpsideDown;
            let frame = CGRectMake(0, 0, min(self.view.frame.size.width,self.view.frame.size.height), max(self.view.frame.size.width,self.view.frame.size.height));
            self.previewLayer.frame = frame
            updateMaskLayer(frame)
        }
    }
    
    override func supportedInterfaceOrientations() -> UIInterfaceOrientationMask {
        return .Portrait
    }
    
    override func prefersStatusBarHidden() -> Bool {
        return false;
    }
    
    func updateMaskLayer(frame: CGRect) {
        
        let sRect = CGRectMake(315, 35, 650, 650);
        let scale:CGFloat = min(CGRectGetWidth(frame), CGRectGetHeight(frame)) / 720.0;
        let size = CGRectGetWidth(sRect) * scale;
        let mRect = CGRectMake(CGRectGetMidX(frame) - size / 2.0 , CGRectGetMidY(frame) - size / 2.0 - frame.origin.y, size, size);
        
        maskLayer.maskRect = mRect;
        maskLayer.frame = frame;
        maskLayer.setNeedsDisplay();
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
}

//MARK: - Init Capture
extension ViewController {
    private func checkCameraAuthorizationStatus() -> (flag: Bool, message: String) {
        let cameraMediaType = AVMediaTypeVideo
        let cameraAuthorizationStatus = AVCaptureDevice.authorizationStatusForMediaType(cameraMediaType)
        
        var flag: Bool = false
        var message = String()
        
        switch cameraAuthorizationStatus {
            
        // The client is authorized to access the hardware supporting a media type.
        case .Authorized:
            flag = true
            // The client is not authorized to access the hardware for the media type. The user cannot change
        // the client's status, possibly due to active restrictions such as parental controls being in place.
        case .Restricted:
            flag = false
            message = "The client is not authorized to access the hardware for the media type. The user cannot change the client's status, possibly due to active restrictions such as parental controls being in place."
            
        case .Denied:
            flag = false
            message = "The user explicitly denied access to the hardware supporting a media type for the client. Go to Settings and enable permission."
        // Indicates that the user has not yet made a choice regarding whether the client can access the hardware.
        case .NotDetermined:
            // Prompting user for the permission to use the camera.
            flag = true
            AVCaptureDevice.requestAccessForMediaType(cameraMediaType) { granted in
                if granted {
                    print("Granted access to \(cameraMediaType)")
                } else {
                    print("Not granted access to \(cameraMediaType)")
                    flag = false
                    message = "The user explicitly denied access to the hardware supporting a media type for the client."
                }
            }
        }
        return (flag: flag, message: message)
    }
    
    private func deinitCapture() {
        if (focusTimer != nil){
            focusTimer?.invalidate();
            focusTimer = nil;
        }
        
        if (captureSession != nil){
            
            captureSession = nil;
            
            previewLayer.removeFromSuperlayer();
            previewLayer = nil;
        }
    }
    
    private func initCapture(){
        
        self.device = AVCaptureDevice.defaultDeviceWithMediaType(AVMediaTypeVideo);
        do {
            
            let captureImput:AVCaptureDeviceInput = try AVCaptureDeviceInput(device: self.device)
            let captureOutput:AVCaptureVideoDataOutput = AVCaptureVideoDataOutput();
            captureOutput.alwaysDiscardsLateVideoFrames = true;
            captureOutput.setSampleBufferDelegate(self, queue: dispatch_get_main_queue());
            
            captureOutput.videoSettings = [kCVPixelBufferPixelFormatTypeKey: Int(kCVPixelFormatType_420YpCbCr8BiPlanarVideoRange)]
            
            self.captureSession = AVCaptureSession();
            
            self.captureSession.addInput(captureImput);
            self.captureSession.addOutput(captureOutput);
            
            
        }catch _{}
        var resX:Int32 = 640
        var resY:Int32 = 480
        
        if self.captureSession.canSetSessionPreset(AVCaptureSessionPreset1280x720){
            print("Set preview port to 1280X720");
            self.captureSession.sessionPreset = AVCaptureSessionPreset1280x720;
            
            resX = 1280
            resY = 720
            
        } else {
            if self.captureSession.canSetSessionPreset(AVCaptureSessionPreset640x480){
                print("Set preview port to 640x480");
                self.captureSession.sessionPreset = AVCaptureSessionPreset640x480;
            }
        }
        
        
        print("Number of processors \(NSProcessInfo.processInfo().processorCount)")
        
        if ( USE_15_FPS /*NSProcessInfo.processInfo().processorCount < 2*/){
            do{
                // Limit camera FPS to 15 for single core devices (iPhone 4 and older) so more CPU power is available for decoder
                try self.device.lockForConfiguration();
                self.device.activeVideoMinFrameDuration = CMTimeMake(1,15)
                self.device .unlockForConfiguration()
                print("activeVideoMinFrameDuration: \(self.device.activeVideoMinFrameDuration)")
            }catch _{}
        }else if USE_60_FPS {
        
            for vFormat in self.device.formats{
                if let vFormat = vFormat  as? AVCaptureDeviceFormat {
                    let description = vFormat.formatDescription
                    let rates = vFormat.videoSupportedFrameRateRanges[0] as! AVFrameRateRange
                    
                    let maxrate = rates.maxFrameRate
                    let minrate = rates.minFrameRate
                    
                    let dimensions: CMVideoDimensions = CMVideoFormatDescriptionGetDimensions(description);
                    if maxrate > 59 && Int32(CMFormatDescriptionGetMediaSubType(description)) == Int32(kCVPixelFormatType_420YpCbCr8BiPlanarVideoRange) && dimensions.width == resX && dimensions.height == resY {
                        do{
                            try self.device.lockForConfiguration()
                            self.device.activeFormat = vFormat
                            self.device.activeVideoMinFrameDuration = CMTimeMake(Int64(10), Int32(minrate * 10))
                            self.device.activeVideoMaxFrameDuration = CMTimeMake(Int64(10), Int32(600))
                            self.device.unlockForConfiguration()
                        }catch _{}
                        
                    }
                }
            }
        }
        
        availableThreads = min(MAX_THREADS, NSProcessInfo.processInfo().processorCount)
        activeThreads = 0
        
        self.previewLayer = AVCaptureVideoPreviewLayer(session: self.captureSession)
        
        if self.interfaceOrientation == UIInterfaceOrientation.LandscapeLeft{
            self.previewLayer.connection.videoOrientation = AVCaptureVideoOrientation.LandscapeLeft;
            self.previewLayer.frame = CGRectMake(0, 0, max(self.view.frame.size.width,self.view.frame.size.height), min(self.view.frame.size.width,self.view.frame.size.height));
        }
        
        if self.interfaceOrientation == UIInterfaceOrientation.LandscapeRight{
            self.previewLayer.connection.videoOrientation = AVCaptureVideoOrientation.LandscapeRight;
            self.previewLayer.frame = CGRectMake(0, 0, max(self.view.frame.size.width,self.view.frame.size.height), min(self.view.frame.size.width,self.view.frame.size.height));
        }
        
        if self.interfaceOrientation == UIInterfaceOrientation.Portrait{
            self.previewLayer.connection.videoOrientation = AVCaptureVideoOrientation.Portrait;
            self.previewLayer.frame = CGRectMake(0, 0, min(self.view.frame.size.width,self.view.frame.size.height), max(self.view.frame.size.width,self.view.frame.size.height));
        }
        
        if self.interfaceOrientation == UIInterfaceOrientation.PortraitUpsideDown{
            self.previewLayer.connection.videoOrientation = AVCaptureVideoOrientation.PortraitUpsideDown;
            self.previewLayer.frame = CGRectMake(0, 0, min(self.view.frame.size.width,self.view.frame.size.height), max(self.view.frame.size.width,self.view.frame.size.height));
        }
        
        
        self.previewLayer.videoGravity = AVLayerVideoGravityResizeAspectFill;
        
        self.view.layer.addSublayer(self.previewLayer);
        
        self.focusTimer = NSTimer.scheduledTimerWithTimeInterval(2.0, target: self, selector: #selector(ViewController.reFocus), userInfo: nil, repeats: true);
        
    }
}


//MARK: - Camera func
extension ViewController {
    override func touchesBegan(touches: Set<UITouch>, withEvent event: UIEvent?) {
        self.toggleTorch()
    }
    
    func reFocus() {
        do{
            
            try self.device?.lockForConfiguration()
            if (self.device.focusPointOfInterestSupported){
                self.device.focusPointOfInterest = CGPointMake(0.49,0.49);
                self.device.focusMode = AVCaptureFocusMode.AutoFocus;
            }
            self.device.unlockForConfiguration();
            
        }catch _{
            
        }
    }
    
    func toggleTorch() {
        if (self.device.isTorchModeSupported(AVCaptureTorchMode.On)) {
            do{
                try self.device?.lockForConfiguration()
                if (self.device.torchMode == AVCaptureTorchMode.On){
                    self.device.torchMode = AVCaptureTorchMode.Off;
                }else {
                    self.device.torchMode = AVCaptureTorchMode.On;
                }
                
                self.device.unlockForConfiguration();
                
            }catch _{
            }
        }
    }
    
    func startScanning() {
        self.captureSession?.startRunning()
        self.previewLayer?.hidden = false
        self.state = cameraState.CAMERA
    }
    
    func stopScanning() {
        self.captureSession?.stopRunning();
        self.state = cameraState.NORMAL
        self.previewLayer?.hidden = true
    }
}

//MARK: - Capture Buffer
extension ViewController {
    func captureOutput(captureOutput: AVCaptureOutput!, didOutputSampleBuffer sampleBuffer: CMSampleBuffer!, fromConnection connection: AVCaptureConnection!) {
        if state != cameraState.CAMERA && state != cameraState.CAMERA_DECODING {
            return;
        }
        
        if self.activeThreads >= self.availableThreads {
            return;
        }
        
        if self.state != cameraState.CAMERA_DECODING {
            self.state = cameraState.CAMERA_DECODING;
        }
        
        self.activeThreads = self.activeThreads+1
        
        let pixelBuffer = CMSampleBufferGetImageBuffer(sampleBuffer);
        
        CVPixelBufferLockBaseAddress(pixelBuffer!,0);
        
        //Get information about the image
        let baseAddress = CVPixelBufferGetBaseAddressOfPlane(pixelBuffer!,0);
        
        let bytesPerRow = CVPixelBufferGetBytesPerRowOfPlane(pixelBuffer!,0);
        let width:Int32 = Int32(bytesPerRow);
        let height:Int32 = Int32(CVPixelBufferGetHeightOfPlane(pixelBuffer!,0));
        
        let frameBuffer  = UnsafeMutablePointer<UInt8>.alloc(Int(width * height));
        frameBuffer.initializeFrom(UnsafeMutablePointer<UInt8>(baseAddress), count: Int(width * height));
        let data = NSData(bytesNoCopy: frameBuffer, length: Int(Int32(bytesPerRow) * height))
        
        CVPixelBufferUnlockBaseAddress(pixelBuffer!, 0);
        
        dispatch_async(dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0)) {
            var resultMessage:String? = ""
            
            if data.length > 0 {
                let widthInt = Int(width)
                let heigthInt = Int(height)
                
                //Recognize Aztec Code
                resultMessage = PSPAztecReader.recognize(data, andWidth: widthInt, andHeight: heigthInt)
            }
            
            if (resultMessage != nil && !resultMessage!.isEmpty && self.state != .DECODE_DISPLAY) {
                dispatch_once(&self.tokenOnce, { [unowned self] () -> Void in
                    self.state = cameraState.DECODE_DISPLAY
                    self.captureSession.stopRunning()
                    
                    dispatch_async(dispatch_get_main_queue()) {
                        self.captureSession.stopRunning();
                        
                        var resultString:String = ""
                        resultString = resultMessage!
                        
                        //use new ios 8 alerts
                        if #available(iOS 8.0, *) {
                            let alertController = UIAlertController(title: "Aztec", message: resultString, preferredStyle:.Alert)
                            let cancelAction = UIAlertAction(title: "Close", style: .Cancel) { action in
                                self.state = cameraState.CAMERA;
                                self.captureSession.startRunning();
                                self.tokenOnce = 0
                            }
                            alertController.addAction(cancelAction)
                            
                            self.presentViewController(alertController, animated: true, completion: nil)
                            
                        } else {
                            //use old alerts
                            let alert = UIAlertView()
                            alert.title = "Aztec";
                            alert.message = resultString;
                            alert.addButtonWithTitle("Close")
                            alert.cancelButtonIndex = 0;
                            alert.delegate = self;
                            
                            alert.show()
                        }
                    }
                    })
            } else {
                self.state = cameraState.CAMERA;
            }
            
            self.activeThreads = self.activeThreads-1
        }
    }
}

//MARK: - UIAlertViewDelegate
extension ViewController: UIAlertViewDelegate {
    func alertView(alertView: UIAlertView, clickedButtonAtIndex buttonIndex: Int) {
        self.captureSession.startRunning()
        self.tokenOnce = 0
        self.state = cameraState.CAMERA
    }
}
